"""django-imdb management command to export IMDB TSV data."""

from __future__ import annotations

import logging
from pathlib import Path

from django.core.management.base import BaseCommand, CommandParser

from django_imdb.export_tsv import export_tsv_files

logger = logging.getLogger(f"django_imdb.{__name__}")


class Command(BaseCommand):
    """Management command to export IMDB TSV data."""

    help = "Export TSV files"

    def add_arguments(self, parser: CommandParser) -> None:
        """Command-line arguments for the manage.py export_imdb_tsv command."""
        parser.add_argument("--export_dir", type=Path, default=Path("~/.cache/django-imdb-tsv-export"))
        parser.add_argument("--skip_name_basics", type=bool, default=False)
        parser.add_argument("--skip_title_basics", type=bool, default=False)
        parser.add_argument("--skip_title_akas", type=bool, default=False)
        parser.add_argument("--skip_title_principals", type=bool, default=False)
        parser.add_argument("--skip_title_episodes", type=bool, default=False)
        parser.add_argument("--skip_title_ratings", type=bool, default=False)

    def handle(self, *args: str, **options: dict[str, str | bool | Path]) -> None:  # noqa: ARG002
        """Export data to TSV files."""
        self.stdout.write(self.style.SUCCESS("Beginning export"))
        logging.basicConfig(level=logging.DEBUG)
        logger.debug(options)
        export_tsv_files(**options)  # type: ignore[arg-type]
        self.stdout.write(self.style.SUCCESS(f"Export done! Files can be found in {options['export_dir']}"))



"""Utils for django-imdb."""

import logging
from pathlib import Path
from typing import TextIO
from urllib.request import Request, urlopen

logger = logging.getLogger(f"django_imdb.{__name__}")
MiB = 1024 * 1024


def count_lines(f: TextIO) -> int:
    """Count lines in a byte iterable. Borrowed and adapted from imdb-sqlite."""
    lines = 0
    chunk = f.read(MiB)
    while chunk:
        lines += chunk.count("\n")
        chunk = f.read(MiB)
    return lines


def download_file(url: str, path: Path) -> None:
    """Download a file."""
    request = Request(url)  # noqa: S310
    logger.debug(f"Downloading {url} to {path} ...")
    with urlopen(request) as response, path.open("wb") as f:  # noqa: S310
        f.write(response.read())


def minsec(seconds: int) -> tuple[int, int]:
    """Return minutes and seconds."""
    spm = 60
    if seconds > spm:
        mins = seconds // spm
        secs = round(seconds % spm)
    else:
        mins = 0
        secs = seconds
    return mins, secs



"""Model definitions for django-imdb."""

import os
from pathlib import Path
from typing import ClassVar

from django.db import models


class TsvBaseModel(models.Model):
    """Abstract model for TSV based model."""

    class Meta:
        """Abstract model."""

        abstract = True

    class TsvMeta:
        """Data structure for the IMDB TSV file this model is based on."""

        filename: Path
        fieldmap: ClassVar[dict[str, tuple[str, type, type]]] = {}
        get_or_create_fks: ClassVar[dict[str, tuple[type[models.Model], str]]]
        unique_fields: ClassVar[list[str]]
        skip_bool: str
        type_casts: ClassVar[dict[str, tuple[type, type]]] = {}

    @property
    def tsvline(self) -> str:
        """Return a string representation of this model instance suitable for export to an IMDB TSV file."""
        elements = []
        for key in type(self).TsvMeta.fieldmap:
            # get value
            value = getattr(self, key)
            if value is not None:
                # cast to correct type
                elements.append(self.TsvMeta.fieldmap[key][2](value))
            else:
                # use imdb nulls
                elements.append("\\N")
        return "\t".join([str(x) for x in elements])


class TitleType(models.Model):
    """The type of a title like movie or tvEpisode."""

    name: models.CharField[str, str] = models.CharField(primary_key=True, help_text="The type.")

    def __str__(self) -> str:
        """Return a string representation of this TitleType."""
        return self.name


class Title(TsvBaseModel):
    """Titles. The basic IMDB unit which everything else relates to."""

    title_id: models.CharField[str, str] = models.CharField(primary_key=True, help_text="The main title ID.")
    title_type: models.ForeignKey[TitleType, TitleType] = models.ForeignKey(
        "TitleType",
        default="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER",
        on_delete=models.PROTECT,
        related_name="titles",
    )
    primary_title: models.CharField[str, str] = models.CharField(
        default="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER",
        help_text="The primary/most used title of the movie",
    )
    original_title: models.CharField[str, str] = models.CharField(
        default="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER",
        help_text="The original title of the movie, in the original language.",
    )
    is_adult: models.BooleanField[bool, bool] = models.BooleanField(default=False)
    premiered: models.IntegerField[int, int] = models.IntegerField(
        blank=True, null=True, help_text="The year of the premiere of this title."
    )
    ended: models.IntegerField[int, int] = models.IntegerField(
        blank=True, null=True, help_text="End year (for TV series only)."
    )
    runtime_minutes: models.IntegerField[int, int] = models.IntegerField(
        blank=True, null=True, help_text="The runtime in minutes of this title."
    )
    genres: models.CharField[str, str] = models.CharField(blank=True, help_text="Up to three genres for this title.")

    class Meta:
        """Set ordering."""

        ordering = ["title_id"]

    class TsvMeta(TsvBaseModel.TsvMeta):
        """Data structure for the TSV file this model is based on."""

        filename = Path("title.basics.tsv.gz")
        fieldmap = {
            "title_id": ("tconst", str, str),
            "title_type_id": ("titleType", str, str),
            "primary_title": ("primaryTitle", str, str),
            "original_title": ("originalTitle", str, str),
            "is_adult": ("isAdult", bool, int),
            "premiered": ("startYear", int, int),
            "ended": ("endYear", int, int),
            "runtime_minutes": ("runtimeMinutes", int, int),
            "genres": ("genres", str, str),
        }
        get_or_create_fks = {"title_type_id": (TitleType, "name")}
        unique_fields = ["title_id"]
        skip_bool = "skip_title_basics"

    def __str__(self) -> str:
        """Return a string representation of this Title."""
        return f"{self.title_id} ({self.title_type}) - {self.primary_title} ({self.premiered})"

    @property
    def imdb_url(self) -> str:
        """Return IMDB url for this title."""
        return f"https://www.imdb.com/title/{self.title_id}/"

    @property
    def genrelist(self) -> list[str]:
        """Return a list of genres."""
        return self.genres.split(",")

    @property
    def yearlist(self) -> list[str]:
        """Return a list of years."""
        if self.premiered and self.ended:
            # this title has an end year
            return [str(x) for x in range(self.premiered, self.ended + 1)]
        if self.premiered:
            # this title only has a premiered year
            return [str(self.premiered)]
        # no years available
        return []

    @property
    def dirname(self) -> str:
        """Return a suitable dirname for this title."""
        return f"{self.primary_title} ({self.premiered})".replace(os.sep, "_")


###################################################################################################


class Person(TsvBaseModel):
    """Human beings. In the IMDB data not everyone has a name, and far from everyone has born/died years."""

    person_id: models.CharField[str, str] = models.CharField(primary_key=True, help_text="The main person ID")
    name: models.CharField[str, str] = models.CharField(
        blank=True,
        default="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER",
        help_text="The name of this person (when known)",
    )
    born: models.IntegerField[int, int] = models.IntegerField(
        blank=True, null=True, help_text="Birth year of this person (when known)"
    )
    died: models.IntegerField[int, int] = models.IntegerField(
        blank=True, null=True, help_text="Death year of this person (when known)"
    )
    primary_professions: models.CharField[str, str] = models.CharField(
        blank=True,
        help_text="Comma-separated list of up to three primary professions for this person.",
    )
    known_for_titles: models.CharField[str, str] = models.CharField(
        blank=True,
        help_text="The titles this person is most known for working on.",
    )

    class Meta:
        """Set ordering."""

        ordering = ["person_id"]

    class TsvMeta(TsvBaseModel.TsvMeta):
        """Data structure for the TSV file this model is based on."""

        filename = Path("name.basics.tsv.gz")
        fieldmap = {
            "person_id": ("nconst", str, str),
            "name": ("primaryName", str, str),
            "born": ("birthYear", int, int),
            "died": ("deathYear", int, int),
            "primary_professions": ("primaryProfession", str, str),
            "known_for_titles": ("knownForTitles", str, str),
        }
        get_or_create_fks = {}
        unique_fields = ["person_id"]
        skip_bool = "skip_name_basics"

    def __str__(self) -> str:
        """Return a string representation of a person."""
        return (
            f"{self.person_id} {self.name} ({self.born if self.born else 'unknown'}-"
            f"{self.died if self.died else 'unknown'})"
        )


###################################################################################################


class AkaType(models.Model):
    """Types of AKAs (alternative titles)."""

    name: models.CharField[str, str] = models.CharField(primary_key=True, help_text="The type name.")

    def __str__(self) -> str:
        """Return a string representation of AkaType."""
        return self.name


class AkaRegion(models.Model):
    """Regions for AKAs (alternative titles)."""

    name: models.CharField[str, str] = models.CharField(primary_key=True, help_text="The region name.")

    def __str__(self) -> str:
        """Return a string representation of AkaRegion."""
        return self.name


class AkaLanguage(models.Model):
    """Languages for AKAs (alternative titles)."""

    name: models.CharField[str, str] = models.CharField(primary_key=True, help_text="The language name.")

    def __str__(self) -> str:
        """Return a string representation of AkaLanguage."""
        return self.name


class Aka(TsvBaseModel):
    """Original and alternative titles. title.akas.tsv.gz."""

    title: models.ForeignKey[Title, Title] = models.ForeignKey("Title", on_delete=models.PROTECT, related_name="akas")
    ordering: models.IntegerField[int, int] = models.IntegerField()
    aka: models.CharField[str, str] = models.CharField(help_text="The title/AKA.")
    region: models.ForeignKey[AkaRegion, AkaRegion] = models.ForeignKey(
        "AkaRegion",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="akas",
        help_text="The region of this AKA (where relevant)",
    )
    language: models.ForeignKey[AkaLanguage, AkaLanguage] = models.ForeignKey(
        "AkaLanguage",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="akas",
        help_text="The language of this AKA (where relevant)",
    )
    aka_type: models.ForeignKey[AkaType, AkaType] = models.ForeignKey(
        "AkaType",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="akas",
        help_text="The type of this AKA",
    )
    attributes: models.CharField[str, str] = models.CharField(
        blank=True, help_text="Additional comma-separated attributes for this title."
    )
    is_original_title: models.BooleanField[bool, bool] = models.BooleanField()

    class Meta:
        """Set ordering."""

        ordering = ["title", "ordering"]

    class TsvMeta(TsvBaseModel.TsvMeta):
        """Data structure for the TSV file this model is based on."""

        filename = Path("title.akas.tsv.gz")
        fieldmap = {
            "title_id": ("titleId", str, str),
            "ordering": ("ordering", int, int),
            "aka": ("title", str, str),
            "region_id": ("region", str, str),
            "language_id": ("language", str, str),
            "aka_type_id": ("types", str, str),
            "attributes": ("attributes", str, str),
            "is_original_title": ("isOriginalTitle", bool, int),
        }
        get_or_create_fks = {
            "title_id": (Title, "title_id"),
            "region_id": (AkaRegion, "name"),
            "language_id": (AkaLanguage, "name"),
            "aka_type_id": (AkaType, "name"),
        }
        unique_fields = ["id"]
        skip_bool = "skip_title_akas"

    def __str__(self) -> str:
        """Return a string representation of an Aka."""
        return (
            f"{self.title.title_type} {self.title.title_id} '{self.title.primary_title}' AKA "
            f"'{self.aka}' (type: {self.aka_type or 'N/A'}, language: {self.language or 'N/A'}, "
            f"region: {self.region or 'N/A'})"
        )


###################################################################################################


class CrewCategory(models.Model):
    """The category of work done by a person on a title (movie/episode)."""

    name: models.CharField[str, str] = models.CharField(primary_key=True, help_text="The category")

    def __str__(self) -> str:
        """Return a string representation of a CrewCategory."""
        return self.name


class Crew(TsvBaseModel):
    """Crew members. title.principals.tsv.gz. Some categories of work has more than one job."""

    title: models.ForeignKey[Title, Title] = models.ForeignKey(
        "Title", on_delete=models.PROTECT, related_name="crewmembers"
    )
    ordering: models.IntegerField[int, int] = models.IntegerField()
    person: models.ForeignKey[Person, Person] = models.ForeignKey(
        "Person", on_delete=models.PROTECT, related_name="crews"
    )
    category: models.ForeignKey[CrewCategory, CrewCategory] = models.ForeignKey(
        "CrewCategory", on_delete=models.PROTECT, related_name="crews"
    )
    job: models.CharField[str, str] = models.CharField(
        blank=True, help_text="The job performed by this crewmember (where relevant)"
    )
    characters: models.CharField[str, str] = models.CharField(
        blank=True, help_text="The characters played by this crewmember (where relevant)"
    )

    class Meta:
        """Set ordering."""

        ordering = ["title", "ordering"]

    class TsvMeta(TsvBaseModel.TsvMeta):
        """Data structure for the TSV file this model is based on."""

        filename = Path("title.principals.tsv.gz")
        fieldmap = {
            "title_id": ("tconst", str, str),
            "ordering": ("ordering", int, int),
            "person_id": ("nconst", str, str),
            "category_id": ("category", str, str),
            "job": ("job", str, str),
            "characters": ("characters", str, str),
        }
        get_or_create_fks = {
            "title_id": (Title, "title_id"),
            "person_id": (Person, "person_id"),
            "category_id": (CrewCategory, "name"),
        }
        unique_fields = ["id"]
        skip_bool = "skip_title_principals"

    def __str__(self) -> str:
        """Return a string representation."""
        played = f"(played {self.clean_characters}) " if self.characters else ""
        return f"{self.person} worked as {self.category} {played}on {self.title}"

    @property
    def clean_characters(self) -> str:
        """Return a version of characters without []."""
        return str(self.characters.replace("[", "").replace("]", ""))


###################################################################################################


class Episode(TsvBaseModel):
    """Episodes of TV shows. Not all shows have seasons and/or episode numbers."""

    show_title: models.ForeignKey[Title, Title] = models.ForeignKey(
        "Title", on_delete=models.PROTECT, related_name="tvshows"
    )
    episode_title: models.ForeignKey[Title, Title] = models.ForeignKey(
        "Title", on_delete=models.PROTECT, related_name="episodes"
    )
    season_number: models.IntegerField[int, int] = models.IntegerField(blank=True, null=True)
    episode_number: models.IntegerField[int, int] = models.IntegerField(blank=True, null=True)

    class Meta:
        """Set ordering."""

        ordering = ["show_title"]

    class TsvMeta(TsvBaseModel.TsvMeta):
        """Data structure for the TSV file this model is based on."""

        filename = Path("title.episode.tsv.gz")
        fieldmap = {
            "show_title_id": ("parentTconst", str, str),
            "episode_title_id": ("tconst", str, str),
            "season_number": ("seasonNumber", int, int),
            "episode_number": ("episodeNumber", int, int),
        }
        get_or_create_fks = {
            "show_title_id": (Title, "title_id"),
            "episode_title_id": (Title, "title_id"),
        }
        unique_fields = ["id"]
        skip_bool = "skip_title_episodes"

    def __str__(self) -> str:
        """Return a string representation of an Episode."""
        return f"Show {self.show_title} episode {self.episode_title}"


###################################################################################################


class Rating(TsvBaseModel):
    """Ratings."""

    title: models.OneToOneField[Title, Title] = models.OneToOneField("Title", on_delete=models.PROTECT)
    rating: models.FloatField[float, float] = models.FloatField()
    votes: models.IntegerField[int, int] = models.IntegerField()

    class Meta:
        """Set ordering."""

        ordering = ["title"]

    class TsvMeta(TsvBaseModel.TsvMeta):
        """Data structure for the TSV file this model is based on."""

        filename = Path("title.ratings.tsv.gz")
        fieldmap = {
            "title_id": ("tconst", str, str),
            "rating": ("averageRating", float, float),
            "votes": ("numVotes", int, int),
        }
        get_or_create_fks = {
            "title_id": (Title, "title_id"),
        }
        unique_fields = ["title"]
        skip_bool = "skip_title_ratings"

    def __str__(self) -> str:
        """Return a string representation of a Rating."""
        return f"{self.title} rating {self.rating} ({self.votes} votes)"



"""Export related functionality."""

from __future__ import annotations

import gzip
import logging
import time
from typing import TYPE_CHECKING, Any

from .models import Aka, Crew, Episode, Person, Rating, Title, TsvBaseModel

if TYPE_CHECKING:
    from pathlib import Path

    from django.db import models

logger = logging.getLogger(f"django_imdb.{__name__}")


def export_tsv_files(  # noqa: PLR0913
    *,
    export_dir: Path,
    skip_name_basics: bool = False,
    skip_title_basics: bool = False,
    skip_title_akas: bool = False,
    skip_title_principals: bool = False,
    skip_title_episodes: bool = False,
    skip_title_ratings: bool = False,
    **kwargs: dict[str, Any],
) -> None:
    """Export IMDB TSV files."""
    export_dir = export_dir.expanduser()
    export_dir.mkdir(parents=True, exist_ok=True)
    for tsvmodel in [Title, Person, Aka, Crew, Episode, Rating]:
        tsv = tsvmodel.TsvMeta
        if locals().get(tsv.skip_bool):
            logger.info(f"Skipping export of {tsv}")
            continue
        path = export_dir / tsv.filename
        objects = tsvmodel.objects.all()
        export_objects(
            export_dir=export_dir,
            model=tsvmodel,
            objects=objects,  # type: ignore[arg-type]
        )


def export_objects(
    export_dir: Path,
    model: type[TsvBaseModel],
    objects: models.QuerySet[Title | Person | Aka | Crew | Episode | Rating],
    batch_size: int = 100000,
) -> None:
    """Export IMDB data."""
    exported = 0
    tsv = model.TsvMeta
    totalstart = time.time()
    p = export_dir / tsv.filename
    logger.debug(f"Exporting file {p} ...")
    with gzip.open(p, "wt") as f:
        headers = "	".join([v[0] for k, v in tsv.fieldmap.items()])
        f.write(f"{headers}\n")
        # batching is only for reporting
        total = objects.count()
        logger.debug(f"Exporting {total} objects to TSV...")
        while exported < total:
            start = time.time()
            for obj in objects[exported : exported + batch_size]:
                if isinstance(obj, Title) and obj.primary_title == "PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER":
                    continue
                if isinstance(obj, Person) and obj.name == "PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER":
                    continue
                f.write(f"{obj.tsvline}\n")
                exported += 1
            duration = time.time() - start
            logger.debug(
                f"Exporting {exported}-{exported + batch_size} of {total} {model} django "
                f"objects took {round(duration, 2)} seconds, {round(batch_size / (duration))}/sec"
            )
    duration = time.time() - totalstart
    logger.info(
        f"Exported {exported} {model} objects in {round(duration, 2)} seconds, {round(exported / duration)}/sec"
    )



"""Initial migrations for django-imdb."""
# Generated by Django 6.0 on 2026-01-18 08:55

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    """Initial migrations for django-imdb."""

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AkaLanguage",
            fields=[
                (
                    "name",
                    models.CharField(
                        help_text="The language name.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AkaRegion",
            fields=[
                (
                    "name",
                    models.CharField(help_text="The region name.", primary_key=True, serialize=False),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AkaType",
            fields=[
                (
                    "name",
                    models.CharField(help_text="The type name.", primary_key=True, serialize=False),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CrewCategory",
            fields=[
                (
                    "name",
                    models.CharField(help_text="The category", primary_key=True, serialize=False),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "person_id",
                    models.CharField(
                        help_text="The main person ID",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True,
                        default="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER",
                        help_text="The name of this person (when known)",
                    ),
                ),
                (
                    "born",
                    models.IntegerField(
                        blank=True,
                        help_text="Birth year of this person (when known)",
                        null=True,
                    ),
                ),
                (
                    "died",
                    models.IntegerField(
                        blank=True,
                        help_text="Death year of this person (when known)",
                        null=True,
                    ),
                ),
                (
                    "primary_professions",
                    models.CharField(
                        blank=True,
                        help_text="Comma-separated list of up to three primary professions for this person.",
                    ),
                ),
                (
                    "known_for_titles",
                    models.CharField(
                        blank=True,
                        help_text="The titles this person is most known for working on.",
                    ),
                ),
            ],
            options={
                "ordering": ["person_id"],
            },
        ),
        migrations.CreateModel(
            name="Title",
            fields=[
                (
                    "title_id",
                    models.CharField(
                        help_text="The main title ID.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "primary_title",
                    models.CharField(
                        default="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER",
                        help_text="The primary/most used title of the movie",
                    ),
                ),
                (
                    "original_title",
                    models.CharField(
                        default="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER",
                        help_text="The original title of the movie, in the original language.",
                    ),
                ),
                ("is_adult", models.BooleanField(default=False)),
                (
                    "premiered",
                    models.IntegerField(
                        blank=True,
                        help_text="The year of the premiere of this title.",
                        null=True,
                    ),
                ),
                (
                    "ended",
                    models.IntegerField(
                        blank=True,
                        help_text="End year (for TV series only).",
                        null=True,
                    ),
                ),
                (
                    "runtime_minutes",
                    models.IntegerField(
                        blank=True,
                        help_text="The runtime in minutes of this title.",
                        null=True,
                    ),
                ),
                (
                    "genres",
                    models.CharField(blank=True, help_text="Up to three genres for this title."),
                ),
            ],
            options={
                "ordering": ["title_id"],
            },
        ),
        migrations.CreateModel(
            name="TitleType",
            fields=[
                (
                    "name",
                    models.CharField(help_text="The type.", primary_key=True, serialize=False),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Rating",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("rating", models.FloatField()),
                ("votes", models.IntegerField()),
                (
                    "title",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="django_imdb.title",
                    ),
                ),
            ],
            options={
                "ordering": ["title"],
            },
        ),
        migrations.CreateModel(
            name="Episode",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("season_number", models.IntegerField(blank=True, null=True)),
                ("episode_number", models.IntegerField(blank=True, null=True)),
                (
                    "episode_title",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="episodes",
                        to="django_imdb.title",
                    ),
                ),
                (
                    "show_title",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="tvshows",
                        to="django_imdb.title",
                    ),
                ),
            ],
            options={
                "ordering": ["show_title"],
            },
        ),
        migrations.CreateModel(
            name="Crew",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("ordering", models.IntegerField()),
                (
                    "job",
                    models.CharField(
                        blank=True,
                        help_text="The job performed by this crewmember (where relevant)",
                    ),
                ),
                (
                    "characters",
                    models.CharField(
                        blank=True,
                        help_text="The characters played by this crewmember (where relevant)",
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="crews",
                        to="django_imdb.crewcategory",
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="crews",
                        to="django_imdb.person",
                    ),
                ),
                (
                    "title",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="crewmembers",
                        to="django_imdb.title",
                    ),
                ),
            ],
            options={
                "ordering": ["title", "ordering"],
            },
        ),
        migrations.CreateModel(
            name="Aka",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("ordering", models.IntegerField()),
                ("aka", models.CharField(help_text="The title/AKA.")),
                (
                    "attributes",
                    models.CharField(
                        blank=True,
                        help_text="Additional comma-separated attributes for this title.",
                    ),
                ),
                ("is_original_title", models.BooleanField()),
                (
                    "language",
                    models.ForeignKey(
                        blank=True,
                        help_text="The language of this AKA (where relevant)",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="akas",
                        to="django_imdb.akalanguage",
                    ),
                ),
                (
                    "region",
                    models.ForeignKey(
                        blank=True,
                        help_text="The region of this AKA (where relevant)",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="akas",
                        to="django_imdb.akaregion",
                    ),
                ),
                (
                    "aka_type",
                    models.ForeignKey(
                        blank=True,
                        help_text="The type of this AKA",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="akas",
                        to="django_imdb.akatype",
                    ),
                ),
                (
                    "title",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="akas",
                        to="django_imdb.title",
                    ),
                ),
            ],
            options={
                "ordering": ["title", "ordering"],
            },
        ),
        migrations.AddField(
            model_name="title",
            name="title_type",
            field=models.ForeignKey(
                default="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER",
                on_delete=django.db.models.deletion.PROTECT,
                related_name="titles",
                to="django_imdb.titletype",
            ),
        ),
    ]



"""Import related functionality."""

from __future__ import annotations

import gzip
import logging
import random
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, TextIO

from .models import Aka, Crew, Episode, Person, Rating, Title, TitleType, TsvBaseModel
from .pocketsearch import reindex_pocketsearch
from .utils import count_lines, download_file

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence

MiB = 1024 * 1024
logger = logging.getLogger(f"django_imdb.{__name__}")


def import_tsv_files(  # noqa: PLR0913
    *,
    download_dir: Path = Path("~/.cache/django-imdb-tsv-data"),
    download_host: str = "datasets.imdbws.com",
    skip_name_basics: bool = False,
    skip_title_basics: bool = False,
    skip_title_akas: bool = False,
    skip_title_principals: bool = False,
    skip_title_episodes: bool = False,
    skip_title_ratings: bool = False,
    max_tsv_age_seconds: int = 86400 * 14,
    **kwargs: dict[str, int | None | bool],
) -> None:
    """Download IMDB TSV files (if needed) and call import_objects() for each TSV file."""
    # create placeholder TitleType
    TitleType.objects.get_or_create(name="PLAYTIME_IMDB_DATA_INCONSISTENCY_PLACEHOLDER")
    download_dir.expanduser().mkdir(parents=True, exist_ok=True)
    for tsvmodel in [Title, Person, Aka, Crew, Episode, Rating]:
        tsv = tsvmodel.TsvMeta
        if locals().get(tsv.skip_bool):
            logger.info(f"Skipping import of {tsv}")
            continue
        path = download_dir / tsv.filename
        if path.exists() and time.time() - path.stat().st_mtime > max_tsv_age_seconds:
            logger.debug(f"TSV file {path} is older than {max_tsv_age_seconds}, delete+re-download...")
            path.unlink()
        if not path.exists():
            url = f"https://{download_host}/{tsv.filename}"
            logger.info(f"Downloading file {url} ...")
            download_file(url=url, path=path)
        import_objects(basedir=download_dir, model=tsvmodel)
    # done, update search index before exiting
    reindex_pocketsearch()


def import_objects(
    basedir: Path,
    model: type[TsvBaseModel],
    batch_size: int = 100000,
) -> None:
    """Read a TSV file and create Django objects for each row."""
    objects: list[TsvBaseModel] = []
    imported = 0
    tsv = model.TsvMeta
    known_fks: dict[str, set[str]] = {fk: set() for fk in tsv.get_or_create_fks}
    totalstart = time.time()
    p = basedir / tsv.filename
    with gzip.open(p, "rt") as f:
        lines = count_lines(f=f)
    logger.info(f"Importing file {p} (file contains {lines} records)...")
    with gzip.open(p, "rt") as f:
        for importcount, row in enumerate(tsvreader(f)):
            if not objects:
                # starting new batch
                logger.debug(f"Creating batch of up to {batch_size} in-memory {model} django objects...")
                start = time.time()
            # make sure needed FKs are created before they are needed
            for fk, (fkmodel, field) in tsv.get_or_create_fks.items():
                rowkey = tsv.fieldmap[fk][0]
                if row[rowkey] and row[rowkey] not in known_fks[fk]:
                    # make sure this FK exists in the database
                    kwargs: dict[str, Any | None] = {field: row[rowkey]}
                    obj, created = fkmodel.objects.get_or_create(**kwargs)
                    if created:
                        logger.debug(f"New value '{row[rowkey]}' for fk {fk} found - created in {fkmodel} in database")
                    known_fks[fk].add(getattr(obj, field))
            # build object kwargs
            kwargs = {}
            for field, (column, import_cast, export_cast) in tsv.fieldmap.items():
                value = row[column]
                if value is None:
                    # this field was \N in the TSV,
                    # use "" for strings, use None for other types (including FKs)
                    kwargs[field] = "" if import_cast is str and field not in tsv.get_or_create_fks else None
                else:
                    kwargs[field] = import_cast(export_cast(value))
            # create object in memory
            obj = model(**kwargs)
            objects.append(obj)
            if len(objects) == batch_size:
                # create batch of objects in DB
                create_objects(objects=objects)
                imported += len(objects)
                duration = time.time() - start
                percent = round((imported / lines) * 100, 2)
                logger.info(
                    f"Imported {percent}% ({importcount} out of {lines} total records). "
                    f"Creating latest batch of {len(objects)} {model} in DB took "
                    f"{round(duration, 2)} seconds, {round(len(objects) / (duration))}/sec"
                )
                objects = []
        if objects:
            # create the last batch
            create_objects(objects=objects)
            imported += len(objects)
    duration = time.time() - totalstart
    logger.info(
        f"Done! Imported or updated {imported} {model} objects in "
        f"{round(duration, 2)} seconds, {round(imported / duration)}/sec"
    )


def tsvreader(f: TextIO) -> Generator[dict[str, str | None]]:
    r"""Yield dicts of lines from IMDB TSV. Interpret \N as None. Inspired by imdb-sqlite."""
    keys = [x.strip() for x in next(f).split("\t")]
    for row in f:
        values = [(x.strip() if x and x != "\\N" else None) for x in row.rstrip().split("\t")]
        yield dict(zip(keys, values, strict=False))


def create_objects(
    objects: Sequence[TsvBaseModel],
    bulk_create_batch_size: int = 1000,
) -> None:
    """Create a bunch of objects using bulk_create."""
    model = type(objects[0])
    tsv = model.TsvMeta
    logger.debug(f"Creating or updating {len(objects)} {model} objects in database")
    start = time.time()
    result = model.objects.bulk_create(
        objs=objects,
        update_conflicts=True,
        update_fields=list(set(tsv.fieldmap.keys()).difference(tsv.unique_fields)),
        unique_fields=tsv.unique_fields,
        batch_size=bulk_create_batch_size,
    )
    duration = time.time() - start
    logger.debug(
        f"Creating {len(objects)} {len(result)} {model} objects in database took "
        f"{round(duration, 2)} seconds, {round(len(objects) / (duration))}/sec"
    )
    rand = random.randint(0, len(objects) - 1)  # noqa: S311
    logger.info(f"Random sample: {objects[rand]}")



"""django-imdb management command to download and import IMDB TSV data."""

from __future__ import annotations

import logging
from pathlib import Path

from django.core.management.base import BaseCommand, CommandParser

from django_imdb.import_tsv import import_tsv_files

logger = logging.getLogger(f"django_imdb.{__name__}")


class Command(BaseCommand):
    """Management command to import IMDB TSV data."""

    help = "Import TSV files"

    def add_arguments(self, parser: CommandParser) -> None:
        """Command-line arguments for the manage.py import_imdb_tsv command."""
        parser.add_argument("--download_dir", type=Path, default=Path("~/.cache/django-imdb-tsv-data"))
        parser.add_argument("--download_host", type=str, default="datasets.imdbws.com")
        parser.add_argument("--skip_name_basics", type=bool, default=False)
        parser.add_argument("--skip_title_basics", type=bool, default=False)
        parser.add_argument("--skip_title_akas", type=bool, default=False)
        parser.add_argument("--skip_title_principals", type=bool, default=False)
        parser.add_argument("--skip_title_episodes", type=bool, default=False)
        parser.add_argument("--skip_title_ratings", type=bool, default=False)
        parser.add_argument("--max_tsv_age_seconds", type=int, default=86400)

    def handle(self, *args: str, **options: dict[str, str | bool | Path]) -> None:  # noqa: ARG002
        """Do the thing."""
        self.stdout.write(self.style.SUCCESS("Beginning import"))
        logging.basicConfig(level=logging.DEBUG)
        logger.debug(options)
        import_tsv_files(**options)  # type: ignore[arg-type]



"""App config for django_imdb."""

from django.apps import AppConfig


class DjangoImdbConfig(AppConfig):
    """App config for django_imdb."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_imdb"



"""Pocketsearch related code."""
import pocketsearch  # type: ignore[import-untyped]
from .models import Aka
from .utils import minsec
import logging
from django.utils import timezone

logger = logging.getLogger(f"django_imdb.{__name__}")

# ignore these characters when indexing, they are not token seperators
IGNORE_CHARS = "'()."

class TitleSearchSchema(pocketsearch.Schema):  # type: ignore[misc]
    """Schema definition for pocketsearch."""

    # search_id is title_id+title combined,
    # it is used to identify unique entries in the index
    search_id = pocketsearch.Text(is_id_field=True)
    title_id = pocketsearch.Int(index=True)
    title = pocketsearch.Text(index=True)
    premiered_year = pocketsearch.Int(index=True)
    ended_year = pocketsearch.Int(index=True)
    rating = pocketsearch.Numeric(index=True)
    votes = pocketsearch.Int(index=True)


def pocketsearch_normalise(string: str) -> str:
    """Normalise a string before adding to pocketsearch index."""
    string = pocketsearch.normalize(value=string)
    for char in IGNORE_CHARS:
        string = string.replace(char, "")
    return string


def reindex_pocketsearch(types: list[str] = ["movie"]) -> None:
    """Reindex pocketsearch title search indexes. Default to doing movies only."""
    akas = Aka.objects.filter(title__title_type__in=types)
    akacount = akas.count()
    indexed = 0
    batch_size = 10000
    while indexed < akacount:
        logger.info(f"Indexed {indexed} of {akacount} titles...")
        with pocketsearch.PocketWriter(
            db_name="/home/user/.cache/playtime-searchindex.db",
            schema=TitleSearchSchema,
            index_name="pocketsearch_titles",
            normalize=pocketsearch_normalise,
        ) as pocket_writer:
            starttime = timezone.now()
            for aka in Aka.objects.filter(title__title_type="movie")[indexed : indexed + batch_size]:
                pocket_writer.insert_or_update(
                    search_id=f"{aka.title.title_id}-{aka.aka}",
                    title_id=aka.title.title_id,
                    title=aka.aka,
                    premiered_year=aka.title.premiered,
                    ended_year=aka.title.ended,
                    rating=aka.title.rating.rating if hasattr(aka.title, "rating") else None,
                    votes=aka.title.rating.votes if hasattr(aka.title, "rating") else None,
                )
                indexed += 1
            duration = timezone.now() - starttime
            speed = round(batch_size / duration.total_seconds())
            etr = round((akacount - indexed) / speed)
            mins, secs = minsec(seconds=etr)
            logger.info(f"Reindex speed is {speed}/sec - ETR {mins} minutes {secs} seconds")



"""Admin modules for django-imdb."""

from django.contrib import admin

from .models import Aka, AkaLanguage, AkaRegion, AkaType, Crew, CrewCategory, Episode, Person, Rating, Title, TitleType


@admin.register(TitleType)
class TitleTypeAdmin(admin.ModelAdmin[TitleType]):
    """Admin class for TitleType."""

    list_display = ["name"]


@admin.register(Title)
class TitleAdmin(admin.ModelAdmin[Title]):
    """Admin class for Title."""

    list_display = [
        "title_id",
        "title_type",
        "primary_title",
        "original_title",
        "is_adult",
        "premiered",
        "ended",
        "runtime_minutes",
        "genres",
    ]
    list_filter = ["title_type", "is_adult"]
    search_fields = ["primary_title", "original_title", "premiered", "ended", "genres"]


###################################################################################################


@admin.register(Person)
class PersonAdmin(admin.ModelAdmin[Person]):
    """Admin class for Person."""

    list_display = ["person_id", "name", "born", "died"]
    search_fields = ["name", "born", "died"]


###################################################################################################


@admin.register(AkaType)
class AkaTypeAdmin(admin.ModelAdmin[AkaType]):
    """Admin class for AkaType."""

    list_display = ["name"]


@admin.register(AkaRegion)
class AkaRegionAdmin(admin.ModelAdmin[AkaRegion]):
    """Admin class for AkaRegion."""

    list_display = ["name"]


@admin.register(AkaLanguage)
class AkaLanguageAdmin(admin.ModelAdmin[AkaLanguage]):
    """Admin class for AkaLanguage."""

    list_display = ["name"]


@admin.register(Aka)
class AkaAdmin(admin.ModelAdmin[Aka]):
    """Admin class for Aka."""

    list_display = ["title", "aka", "region", "language", "aka_type", "attributes", "is_original_title"]
    list_filter = ["region", "language"]


###################################################################################################


@admin.register(CrewCategory)
class CrewCategoryAdmin(admin.ModelAdmin[CrewCategory]):
    """Admin class for CrewCategory."""

    list_display = ["name"]


@admin.register(Crew)
class CrewAdmin(admin.ModelAdmin[Crew]):
    """Admin class for Crew."""

    list_display = ["id", "title", "person", "category", "job", "characters"]
    list_filter = ["category"]


###################################################################################################


@admin.register(Episode)
class EpisodeAdmin(admin.ModelAdmin[Episode]):
    """Admin class for Episode."""

    list_display = ["show_title", "episode_title", "season_number", "episode_number"]


###################################################################################################


@admin.register(Rating)
class RatingAdmin(admin.ModelAdmin[Rating]):
    """Admin class for Rating."""

    list_display = ["title_id", "rating", "votes"]
    list_filter = ["title_id"]
